/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.datavec.api.io;

import java.io.*;

public class DataOutputBuffer extends DataOutputStream {

    private static class Buffer extends ByteArrayOutputStream {
        public byte[] getData() {
            return buf;
        }

        public int getLength() {
            return count;
        }

        public Buffer() {
            super();
        }

        public Buffer(int size) {
            super(size);
        }

        public void write(DataInput in, int len) throws IOException {
            int newcount = count + len;
            if (newcount > buf.length) {
                byte newbuf[] = new byte[Math.max(buf.length << 1, newcount)];
                System.arraycopy(buf, 0, newbuf, 0, count);
                buf = newbuf;
            }
            in.readFully(buf, count, len);
            count = newcount;
        }
    }

    private Buffer buffer;

    /** Constructs a new empty buffer. */
    public DataOutputBuffer() {
        this(new Buffer());
    }

    public DataOutputBuffer(int size) {
        this(new Buffer(size));
    }

    private DataOutputBuffer(Buffer buffer) {
        super(buffer);
        this.buffer = buffer;
    }

    /** Returns the current contents of the buffer.
     *  Data is only valid to {@link #getLength()}.
     */
    public byte[] getData() {
        return buffer.getData();
    }

    /** Returns the length of the valid data currently in the buffer. */
    public int getLength() {
        return buffer.getLength();
    }

    /** Resets the buffer to empty. */
    public DataOutputBuffer reset() {
        this.written = 0;
        buffer.reset();
        return this;
    }

    /** Writes bytes from a DataInput directly into the buffer. */
    public void write(DataInput in, int length) throws IOException {
        buffer.write(in, length);
    }

    /** Write to a file stream */
    public void writeTo(OutputStream out) throws IOException {
        buffer.writeTo(out);
    }
}
